from __future__ import annotations

from collections.abc import Hashable
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Any, Callable, TypeVar
from weakref import WeakValueDictionary

import discord
from discord.ext import commands

from .db import AbstractConnection, AbstractDbFactory

T = TypeVar("T")


class Temporaries:
    def __init__(self) -> None:
        self.__values: WeakValueDictionary[Hashable, Any] = WeakValueDictionary()

    def get(self, key: Hashable, factory: Callable[[], T]) -> T | Any:
        return self.__values.get(key) or self.__values.setdefault(key, factory())


class StarState:
    def __init__(self, connection: AbstractConnection) -> None:
        self.connection = connection
        self.temporaries = Temporaries()


@asynccontextmanager
async def state_manager(dbf: AbstractDbFactory):
    async with dbf.from_path(Path("/app/data/stars.db")) as connection:
        yield StarState(connection)


class StarBot(commands.Bot):
    def __init__(self, state: StarState) -> None:
        super().__init__(
            command_prefix="⭐",
            intents=discord.Intents(message_content=True, guild_messages=True, guild_reactions=True, guilds=True),
        )
        self.starstate: StarState = state
